/*
 ****************************************************************************
 * Copyright 2001-2003 by NeTraverse, Inc.
 * This software is distributed under the terms of the GPL
 * which is supplied in the LICENSE file with this distribution
 ***************************************************************************
 * $Id: mki-main.c,v 1.11 2004/10/28 20:02:27 lreiter Exp $
 ***************************************************************************
 */
#include <linux/config.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/version.h>
#include <linux/sched.h>
#include <linux/kdev_t.h>
#include <linux/errno.h>
#include <linux/mm.h>
#include <linux/vmalloc.h>
#include <linux/slab.h>
#include <linux/smp_lock.h>
#include <linux/poll.h>
#include <linux/interrupt.h>
#include <linux/proc_fs.h>
#include <linux/stat.h>
#include <linux/vermagic.h>
#include <asm/mki.h>
#include <asm/uaccess.h>
#include <asm/mman.h>

#include "mkifunc.h"
#include "mki-adapter26.h"

static const char mkia_version[] = "1.3.8";

#ifndef MODULE
#error "mki-adapter must be done as a module"
#endif

#ifdef MODULE_LICENSE
MODULE_LICENSE("GPL");
MODULE_AUTHOR("NeTraverse");
MODULE_DESCRIPTION("NeTraverse MKI Adapter");
#endif

int
mkia_process_owns_fpu(void *vm86p)
{
	struct thread_info *thread = current_thread_info();
	return ((thread->status & TS_USEDFPU) != 0);
}
EXPORT_SYMBOL(mkia_process_owns_fpu);

int 
mkia_register_chrdev(int maj, 
		    char *name, 
		    void *fops)
{
	return register_chrdev(maj, name, (struct file_operations *) fops);
}
EXPORT_SYMBOL(mkia_register_chrdev);

void
mkia_unregister_chrdev(int major, char * name)
{
	unregister_chrdev(major, name);
}
EXPORT_SYMBOL(mkia_unregister_chrdev);


/*
 * populate a mergevma struct from a vmarea
 */
int
mkia_populate_mrgvma(struct mergevma *mrgvma, void * vmap)
{
	mrgvma->offset =
	    (((struct vm_area_struct *) vmap)->vm_pgoff << PAGE_SHIFT);

	mrgvma->vm_start = ((struct vm_area_struct *) vmap)->vm_start;
	mrgvma->vm_end = ((struct vm_area_struct *) vmap)->vm_end;
	mrgvma->nbytes = 
		(((mrgvma->vm_end - mrgvma->vm_start) + 
		  PAGE_SIZE - 1) & PAGE_SIZE);
	mrgvma->npages = mrgvma->nbytes >> PAGE_SHIFT;
	mrgvma->page_size = PAGE_SIZE;
	return 0;
}
EXPORT_SYMBOL(mkia_populate_mrgvma);

int
mkia_remap_page_range(unsigned long user_addr, 
			     unsigned long phys_addr, 
			     unsigned long size,
			     void *vma) 
{
	return remap_page_range(
			(struct vm_area_struct *) vma,
			user_addr, 
			phys_addr, 
			size, 
			((struct vm_area_struct *) vma)->vm_page_prot);
}
EXPORT_SYMBOL(mkia_remap_page_range);

void *
mkia_alloc_file_ops(void *read, 
			   void *write, 
			   void *readdir, 
			   void *poll, 
			   void *ioctl, 
			   void *mmap, 
			   void *open, 
			   void *release)
{
	struct file_operations *ret=kmalloc(sizeof(struct file_operations), 
					    GFP_ATOMIC);
	if (!ret)
		return ret;

	memset(ret, 0, sizeof(struct file_operations));

	ret->read = read;
	ret->write = write;
	ret->readdir = readdir;
	ret->poll = poll;
	ret->ioctl = ioctl;
	ret->mmap = mmap;
	ret->open = open;
	ret->release = release;
	return ret;
}
EXPORT_SYMBOL(mkia_alloc_file_ops);

void
mkia_free_file_ops(void * fops)
{
	kfree(fops);
}
EXPORT_SYMBOL(mkia_free_file_ops);

int 
mkia_get_inode_minor(void *inode)
{
	return MINOR(((struct inode *) inode)->i_rdev);
}
EXPORT_SYMBOL(mkia_get_inode_minor);

int 
mkia_get_file_minor(void *file)
{
	return MINOR(((struct file *)file)->f_dentry->d_inode->i_rdev);
}
EXPORT_SYMBOL(mkia_get_file_minor);

void *
mkia_get_free_pages(int pages, int flags)
{
	int fpflags=0;
	unsigned long ret;
	if (flags & MKIA_NOSLEEP) fpflags |= GFP_ATOMIC;
	if (flags & MKIA_SLEEP) fpflags |= GFP_KERNEL;
	if (flags & MKIA_DMA) fpflags |= GFP_DMA;
	ret = __get_free_pages(fpflags, pages);
	if (ret && (flags & MKIA_ZERO))
		memset((void *) ret, 0, PAGE_SIZE * (pages + 1));
	return (void *) ret;
}
EXPORT_SYMBOL(mkia_get_free_pages);

void 
mkia_free_pages(unsigned long addr, int size)
{
	free_pages(addr, size);
}
EXPORT_SYMBOL(mkia_free_pages);

#ifdef CONFIG_PREEMPT
/*
 * This is set to 0 on the first call to mkia_getparm if the drivers
 * know about preempt
 */
int mkia_disable_kmalloc = 1;
#else
int mkia_disable_kmalloc = 0;
#endif

void * 
mkia_kmalloc(int size, int flags)
{
	int fpflags;
	unsigned long ret;

	fpflags=0;
	if (mkia_disable_kmalloc) {
		printk("mki-adapter: trying to run a non-preemption "
			"capable Win4Lin on a preemptible kernel\n");
		return NULL;
	}

	if (flags & MKIA_NOSLEEP) fpflags |= GFP_ATOMIC;
	if (flags & MKIA_SLEEP) fpflags |= GFP_KERNEL;
	if (flags & MKIA_DMA) fpflags |= GFP_DMA;
	ret = (unsigned long) kmalloc(size, fpflags);
	if (ret && (flags & MKIA_ZERO))
		memset((void *) ret, 0, size);
	return (void *) ret;
}
EXPORT_SYMBOL(mkia_kmalloc);

int
mkia_signal_pending(void *foo)
{
	return signal_pending((struct task_struct *) foo);
}

void * 
mkia_current()
{
	return current;
}


void
mkia_mod_inc_use_count(void * module) 
{
	(void) try_module_get(module);
}
EXPORT_SYMBOL(mkia_mod_inc_use_count);

void
mkia_mod_dec_use_count(void * module) 
{
	(void) module_put(module);
}
EXPORT_SYMBOL(mkia_mod_dec_use_count);

int
mkia_get_current_pid()
{
	return current->pid;
}
EXPORT_SYMBOL(mkia_get_current_pid);

void *
mkia_get_current_task()
{
	return current;
}
EXPORT_SYMBOL(mkia_get_current_task);

void
mkia_kill_proc(int procref, int sig, int foo)
{
	kill_proc(procref, sig, foo);
}
EXPORT_SYMBOL(mkia_kill_proc);

void
mkia_file_set_private_data(void * filp, void *data)
{
    ((struct file *) filp)->private_data = data;
}
EXPORT_SYMBOL(mkia_file_set_private_data);

void *
mkia_file_get_private_data(void * filp)
{
	return ((struct file *) filp)->private_data;
}
EXPORT_SYMBOL(mkia_file_get_private_data);

void *
mkia_file_get_f_pos_addr(void * filp)
{
	return &(((struct file *) filp)->f_pos);
}
EXPORT_SYMBOL(mkia_file_get_f_pos_addr);

long
mkia_file_get_f_mode(void * filp)
{
	return ((struct file *) filp)->f_mode;
}
EXPORT_SYMBOL(mkia_file_get_f_mode);

int
mkia_is_file_nonblock(void *file)
{
	return ((((struct file *)file)->f_flags & O_NONBLOCK) != 0);
}
EXPORT_SYMBOL(mkia_is_file_nonblock);

int 
mkia_copy_to_user(void *dest, void *src, int len)
{
	return copy_to_user(dest, src, len);
}
EXPORT_SYMBOL(mkia_copy_to_user);

int 
mkia_copy_from_user(void *dest, void *src, int len)
{
	return copy_from_user(dest, src, len);
}
EXPORT_SYMBOL(mkia_copy_from_user);

int
mkia_request_irq(int irq,
		void (*handler)(int, void *, struct pt_regs *), 
		unsigned long foo, char *bar, void *baz)
{
	return request_irq(irq, (void *)handler, foo, bar, baz);
}
EXPORT_SYMBOL(mkia_request_irq);

void
mkia_free_irq(int irq, void *baz)
{
	free_irq(irq, baz);
}
EXPORT_SYMBOL(mkia_free_irq);

void  
mkia_kfree(void *pointer)
{
	kfree(pointer);
}
EXPORT_SYMBOL(mkia_kfree);

void * 
mkia_vmalloc(int size)
{
	return vmalloc(size);
}
EXPORT_SYMBOL(mkia_vmalloc);

void  
mkia_vfree(void *pointer)
{
	vfree(pointer);
}
EXPORT_SYMBOL(mkia_vfree);


int
mkia_get_user(void * ptr, int len)
{
	if (len == sizeof(unsigned char)) {
		unsigned char flags;
		if (get_user(flags, (unsigned char *) ptr))
			return -1;
		else
			return flags;
	}
	if (len == sizeof(unsigned short)) {
		unsigned short flags;
		if (get_user(flags, (unsigned short *) ptr))
			return -1;
		else
			return flags;
	}
	if (len == sizeof(unsigned long)) {
		unsigned long flags;
		if (get_user(flags, (unsigned long *) ptr))
			return -1;
		else
			return flags;
	}
	return 0;
}
EXPORT_SYMBOL(mkia_get_user);

int
mkia_put_user(unsigned long flags, void * ptr, int len)
{
	if (len == 4) return put_user((unsigned long) flags, 
			(unsigned long *) ptr);
	if (len == 2) return put_user((unsigned short) flags, 
			(unsigned short *) ptr);
	if (len == 1) return put_user((unsigned char) flags, 
			(unsigned char *) ptr);
	return 0;
}
EXPORT_SYMBOL(mkia_put_user);

int
mkia_strncmp(char *s1, char *s2, int len)
{
	return strncmp(s1,s2,len);
}

int
mkia_strcmp(char *s1, char *s2)
{
	return strcmp(s1,s2);
}

void *
mkia_strncpy(char *s1, char *s2, int len)
{
	return strncpy(s1,s2,len);
}
EXPORT_SYMBOL(mkia_strncpy);

void *
mkia_strcpy(char *s1, char *s2)
{
	return strcpy(s1,s2);
}

void 
mkia_memset(void *mem, int val, int len)
{
	memset(mem, val, len);
}
EXPORT_SYMBOL(mkia_memset);

void 
mkia_memcpy(void *mem, void *mem2, int len)
{
	memcpy(mem, mem2, len);
}
EXPORT_SYMBOL(mkia_memcpy);

void *
mkia_alloc_waitqueuep(void)
{
	void *wq = NULL;

	wq = kmalloc(sizeof(wait_queue_head_t), GFP_ATOMIC);
	if (!wq)
		return NULL;
	memset(wq, 0, sizeof(wait_queue_head_t));
	init_waitqueue_head((wait_queue_head_t *) wq);
	return wq;
}
EXPORT_SYMBOL(mkia_alloc_waitqueuep);

void
mkia_free_waitqueuep(void *waitqp)
{
	kfree(waitqp);
}
EXPORT_SYMBOL(mkia_free_waitqueuep);

rwlock_t _mkia_waitqueue_lock = RW_LOCK_UNLOCKED;

void 
mkia_sleep_on(void *wqp)
{
	sleep_on((wait_queue_head_t *) wqp);
}

void 
mkia_interruptible_sleep_on(void *wqp)
{
	interruptible_sleep_on((wait_queue_head_t *) wqp);
}


void
mkia_poll_wait(void *file, void *wqp, void *wait)
{
	poll_wait((struct file *) file, (wait_queue_head_t *) wqp, 
		  (poll_table *) wait);
}
EXPORT_SYMBOL(mkia_poll_wait);

void
mkia_poll_wake(void *wqp)
{
	wake_up_interruptible((wait_queue_head_t *) wqp);
}
EXPORT_SYMBOL(mkia_poll_wake);

/*
 * Routine to unlock the kernel in mrgioctl for some kernel versions since
 * we use finer locking than linux does and we need the kernel lock to
 * be free.
 */
void *
mkia_mrgioctl_unlock_kernel(void) 
{
#ifdef CONFIG_SMP
	if (current->lock_depth > 0) {
		unlock_kernel();
		return (void *) 1;
	}
#endif
	return NULL;
}
EXPORT_SYMBOL(mkia_mrgioctl_unlock_kernel);

/*
 * Routine to re-lock the kernel in mrgioctl for 2.4 after the ioctl so
 * that linux is happy
 */
void 
mkia_mrgioctl_lock_kernel(void * cookie)
{
#ifdef CONFIG_SMP
	if (cookie)
		lock_kernel();
#endif
}

void
install_mhi_hooks(void)
{
	mhi_hooks[MKI_HOOK_RET_USER] = &mhia_ret_user;
	mhi_hooks[MKI_HOOK_SWITCH_TO] = &mhia_switch_to;
	mhi_hooks[MKI_HOOK_SWITCH_AWAY] = &mhia_switch_away;
	mhi_hooks[MKI_HOOK_EXIT] = &mhia_exit;
	mhi_hooks[MKI_HOOK_SWAP] = &mhia_swap;
}

void
remove_mhi_hooks(void)
{
	mhi_hooks[MKI_HOOK_RET_USER] = &mhi_void_hook;
	mhi_hooks[MKI_HOOK_SWITCH_TO] = &mhi_void_hook;
	mhi_hooks[MKI_HOOK_SWITCH_AWAY] = &mhi_void_hook;
	mhi_hooks[MKI_HOOK_EXIT] = &mhi_void_hook;
	mhi_hooks[MKI_HOOK_SWAP] = &mhi_void_hook;
}


#ifdef CONFIG_PROC_FS

char *mki_adapter_name = "mki-adapter";

struct mkia_proc_string_info {
    char mpsi_name[32];
    struct proc_dir_entry *mpsi_proc_entry;
    char mpsi_string[120];
    int  mpsi_strlen;
} mkia_proc[] = {
	{ "mod-struct-info", 0, "", 0 },
#define MKIA_PROC_MOD_STRUCT_INFO 0
	{ "mod-version-info", 0, "", 0 },
#define MKIA_PROC_MOD_VERSION_INFO 1
	{ "version", 0, "", 0 },
#define MKIA_PROC_VERSION 2
};

#define NUM_MKIA_PROC (sizeof(mkia_proc) / sizeof(mkia_proc[0]))

int mkia_proc_string_read(struct file *file, char __user *buf,
    size_t size, loff_t *ppos)
{
	struct mkia_proc_string_info *mpsip;
	int count;
	int index;
	int err;

	mpsip = (struct mkia_proc_string_info *) (file->private_data);

	/* NOTE: removed ppos test to make things run on 2.6.8 kernels */
	
	index = (int) *ppos;
	if (index > mpsip->mpsi_strlen)
		return 0;

	count = mpsip->mpsi_strlen - index;
	if (count > (int) size)
		count = (int) size;

	err = copy_to_user(buf, &(mpsip->mpsi_string[index]), count);
	if (err)
		return err;

	*ppos += count;
	return count;
}

int
mkia_proc_struct_info_open(struct inode *inode, struct file *file)
{
	struct mkia_proc_string_info *msip;
	msip = &mkia_proc[MKIA_PROC_MOD_STRUCT_INFO];
	file->private_data = (void *) msip;
	return 0;
}

static struct file_operations mkia_struct_info_fops = {
	.owner = THIS_MODULE,
	.open = mkia_proc_struct_info_open,
	.read = mkia_proc_string_read,
};

int
mkia_proc_version_info_open(struct inode *inode, struct file *file)
{
	struct mkia_proc_string_info *msip;
	msip = &mkia_proc[MKIA_PROC_MOD_VERSION_INFO];
	file->private_data = (void *) msip;
	return 0;
}

static struct file_operations mkia_version_info_fops = {
	.owner = THIS_MODULE,
	.open = mkia_proc_version_info_open,
	.read = mkia_proc_string_read,
};

int
mkia_proc_version_open(struct inode *inode, struct file *file)
{
	struct mkia_proc_string_info *msip;
	msip = &mkia_proc[MKIA_PROC_VERSION];
	file->private_data = (void *) msip;
	return 0;
};

static struct file_operations mkia_version_fops = {
	.owner = THIS_MODULE,
	.open = mkia_proc_version_open,
	.read = mkia_proc_string_read,
};

struct proc_dir_entry *mkia_proc_root = NULL;

#endif /* CONFIG_PROC_FS */

void cleanup_module(void);

int 
init_module(void)
{
	install_mhi_hooks();

#ifdef CONFIG_PROC_FS
{
	struct mkia_proc_string_info *mpsip;
	int i;

	mkia_proc_root = create_proc_entry(mki_adapter_name,
		S_IFDIR | 0555, NULL);

	if (mkia_proc_root == NULL)
		goto init_module_fail;

	for (i = 0, mpsip = mkia_proc; i < NUM_MKIA_PROC; i++, mpsip++) {
	    mpsip->mpsi_proc_entry = create_proc_entry(mpsip->mpsi_name,
			S_IFREG | 0444, mkia_proc_root);
	    if (mpsip->mpsi_proc_entry == NULL)
		    goto init_module_fail;
	    switch (i) {
		case MKIA_PROC_MOD_STRUCT_INFO:
		    snprintf(mpsip->mpsi_string, sizeof(mpsip->mpsi_string),
			"4 0x%x 0x%x 0x%x 0x%x\n", sizeof(struct module),
			(int) &(((struct module *) 0)->name[0]),
			(int) &(((struct module *) 0)->init),
#ifdef CONFIG_MODULE_UNLOAD
			(int) &(((struct module *) 0)->exit)
#else
			-1
#endif
		    );
		    mpsip->mpsi_proc_entry->proc_fops = &mkia_struct_info_fops;
		    break;

		case MKIA_PROC_MOD_VERSION_INFO:
		    snprintf(mpsip->mpsi_string, sizeof(mpsip->mpsi_string),
		        "%s\n", VERMAGIC_STRING);
		    mpsip->mpsi_string[sizeof(mpsip->mpsi_string) - 1] = 0;
		    mpsip->mpsi_proc_entry->proc_fops = &mkia_version_info_fops;
		    break;

		case MKIA_PROC_VERSION:
		    snprintf(mpsip->mpsi_string, sizeof(mpsip->mpsi_string),
			"%s\n", mkia_version);
		    mpsip->mpsi_string[sizeof(mpsip->mpsi_string) - 1] = 0;
		    mpsip->mpsi_proc_entry->proc_fops = &mkia_version_fops;
		    break;

		default:
		    printk("Win4Lin: init: unknown procinfo index\n");
		    goto init_module_fail;
	    }
	    mpsip->mpsi_strlen = strlen(mpsip->mpsi_string);
	}
}
#endif /* CONFIG_PROC_FS */

	return 0;


init_module_fail:
	printk("mki-adapter: init_module failed\n");
	cleanup_module();
	return -EINVAL;
}

void 
cleanup_module(void)
{
#ifdef CONFIG_PROC_FS
{
	struct mkia_proc_string_info *mpsip;
	int i;

	for (i = 0, mpsip = mkia_proc; i < NUM_MKIA_PROC; i++, mpsip++) {
		if (mpsip->mpsi_proc_entry == NULL)
			continue;
		remove_proc_entry(mpsip->mpsi_name, mkia_proc_root);
		mpsip->mpsi_proc_entry = NULL;
	}
	if (mkia_proc_root) {
		remove_proc_entry(mki_adapter_name, NULL);
		mkia_proc_root = 0;
	}
}
#endif /* CONFIG_PROC_FS */
	remove_mhi_hooks();
}
