/*
 ****************************************************************************
 * Copyright 2003 by NeTraverse, Inc.
 * This software is distributed under the terms of the GPL
 * which is supplied in the LICENSE file with this distribution
 ***************************************************************************
 * $Id: mki26.c,v 1.13 2004/10/28 21:36:43 lreiter Exp $
 ***************************************************************************
 */
#include <linux/config.h>
#include <linux/version.h>
#include <linux/compiler.h>
#include <linux/types.h>
#include <linux/sched.h>
#include <linux/interrupt.h>
#include <linux/vmalloc.h>
#include <linux/highmem.h>
#include <linux/smp_lock.h>
#include <linux/smp.h>
#include <linux/stddef.h>
#include <linux/file.h>
#include <linux/swap.h>
#include <linux/slab.h>
#include <linux/module.h>
#include <linux/mm.h>
#include <linux/suspend.h>

#include <asm/processor.h>
#include <asm/ptrace.h>
#include <asm/pgtable.h>
#include <asm/desc.h>
#include <asm/hardirq.h>
#include <asm/uaccess.h>
#include <asm/ldt.h>
#include <asm/desc.h>
#include <asm/pgalloc.h>
#include <asm/io.h>
#include <asm/mman.h>
#include <asm/tlbflush.h>
#include <asm/percpu.h>

#include <asm/mki.h>

#include "mkifunc.h"
#include "mki-adapter26.h"

extern struct desc_struct idt_table[];

/* externals */
extern asmlinkage int sys_ftruncate(int, unsigned long);
extern asmlinkage long sys_munmap(unsigned long, size_t);
extern asmlinkage long sys_mprotect(unsigned long, size_t, unsigned long);

atomic_t mkia_context_rss = ATOMIC_INIT(0);

/* Performance/debugging counters */
unsigned int mkia_cnt_rpr_pagereserved = 0;
unsigned int mkia_cnt_rpr_not_dirty_acc = 0;
unsigned int mkia_cnt_rpr_dirty = 0;
unsigned int mkia_cnt_rpr_accessed = 0;

struct idt_gdt_desc_struct {
	unsigned short limit;
	struct desc_struct __attribute__((packed)) *table;
	unsigned short pad;     /* Align for stack variables */
};

struct idt_gdt_desc_struct mkia_idt_ = {(256 * 8) - 1, 0};
#define mkia_idt mkia_idt_.table


static int
mhia_void(void *parm)
{
	return -1;
}

int (*mhia_table[])(void *) = {
	&mhia_void,	     /* SWITCH_TO */
	&mhia_void,	     /* SWITCH_AWAY */
	&mhia_void,	     /* THREAD_EXIT */
	&mhia_void,	     /* RET_USER */
	&mhia_void,	     /* SIGNAL */
	&mhia_void,	     /* QUERY */
	&mhia_void,	     /* SWAP_PAGES */
};

/* The next one might as well go into bss */
static struct desc_struct * mkia_idt_rw;

/*
 * Used in the implementation of the MKI functions
 */
void mkia_assert_failure(char *exprstr, char *filename, int linenum);

#define	MKIA_ASSERT(expression) do { \
	if (unlikely(!(expression))) \
		mkia_assert_failure(#expression, __FILE__, __LINE__); \
} while (0)


#define TSS_SIZE_32 104    	     /* TSS is 104 bytes long for 32 bit TSS */
#define MKI_TSS_IOBITMAP_BITS 1024   /* This many bits in the TSS's iobitmap */
#define MKI_TSS_FUDGE 8 	     /* Fudge factor to add to the tss limit */
#define MKI_TSS_LIMIT(x)     (TSS_SIZE_32 + ((x) >> 3) + MKI_TSS_FUDGE)

/*
 * This is the address where the MKI expects the Linux kernel to live.
 * If someone makes a kernel with PAGE_OFFSET at a different address,
 * then we probably won't work, because the Windows address space
 * is very tight as it is!
 */

#if __PAGE_OFFSET < MKI_END_USER_ADDR
	#error MKI will not work if __PAGE_OFFSET is not >= MKI_END_USER_ADDR
#endif

void
mkia_init(void)
{
}

mkia_task_t * mkia_alloc_task_info(void);
void mkia_pgvfree(void *, unsigned long);
void mkia_alloc_tss(void *, int);

/***************************************************************************/

#ifdef CONFIG_PREEMPT
#ifdef PREEMPT_DEBUG
unsigned long mkia_preempt_enable_counter = 0;
unsigned long mkia_preempt_enable_nosched_counter = 0;
unsigned long mkia_preempt_disable_counter = 0;
#define PREEMPT_COUNT(a) a++;
#else
#define PREEMPT_COUNT(a)
#endif
#endif

void
mkia_preempt_disable(void)
{
#ifdef CONFIG_PREEMPT
	PREEMPT_COUNT(mkia_preempt_disable_counter);
	preempt_disable();
#endif
}

void
mkia_preempt_enable(void)
{
#ifdef CONFIG_PREEMPT
	PREEMPT_COUNT(mkia_preempt_enable_counter);
	preempt_enable();
#endif
}

void
mkia_preempt_enable_nosched(void)
{
#ifdef CONFIG_PREEMPT
	PREEMPT_COUNT(mkia_preempt_enable_nosched_counter);
	preempt_enable_no_resched();
#endif
}

int
mkia_get_preempt_count(void)
{
#ifdef CONFIG_PREEMPT
	return preempt_count();
#else
	return 0;
#endif
}


/*
 * mkia_get_task_info
 * allocates the task_info struct if it's null in the current task
 */
static inline mkia_task_t *
mkia_get_task_info(struct task_struct *curr)
{
	mkia_task_t *mtip;
	if ((mtip = curr->mki_task_info) == NULL) {
		mtip = mkia_alloc_task_info();
		curr->mki_task_info = mtip;
	}
	return mtip;
}

static inline void
mkia_free_ldt(mkia_task_t *mtip)
{
	if (mtip->mti_merge_ldtp != NULL) {
		vfree(mtip->mti_merge_ldtp);
		mtip->mti_merge_ldtp = NULL;
		mtip->mti_save_ldtp = NULL;
	}
}

static inline void
mkia_free_tss(mkia_task_t *mtip)
{
	if (mtip->mti_merge_tssp != NULL) {
		kfree(mtip->mti_merge_tssp);
		mtip->mti_merge_tssp = NULL;
	}
}

static inline void
mkia_free_gdt(mkia_task_t *mtip)
{
	if (mtip->mti_merge_gdtp != NULL) {
		free_page((unsigned long) mtip->mti_merge_gdtp);
		mtip->mti_merge_gdtp = NULL;
	}
}

/* Called from mkia_remove_hook */
static inline void
mkia_cleanup_idt(void)
{
	lock_kernel();
	if (mkia_idt_rw != NULL) {
		if (boot_cpu_data.f00f_bug) {
			mkia_pgvfree(mkia_idt_rw, 2*PAGE_SIZE);
		}
		else {
			free_page((unsigned long) mkia_idt_rw);
		}
		mkia_idt_rw = NULL;
	}
	mkia_idt = NULL;
	unlock_kernel();
}

/*
 * mkia_load_linux_descriptors
 *
 * switch in the linux gdt, idt, ldt, and tr
 */
static inline void
mkia_load_linux_descriptors(void)
{
	int smp_id = smp_processor_id();

	/*
	 * KERNEL gdt elements at the tail end because Linux has
	 * per-CPU GDTs in this version.
	 */
	struct desc_struct *gdttable;

	asm volatile("lgdt %0": : "m" (cpu_gdt_descr[smp_id]));
	asm volatile("lidt %0": : "m" (idt_descr));
	load_LDT_desc();

	/* Clear busy bit prior to ltr */
	gdttable = (struct desc_struct *) (cpu_gdt_descr[smp_id].address);
	gdttable[GDT_ENTRY_TSS].b &= 0xfffffdff;
	load_TR_desc();
}

static inline 
void
mkia_alloc_gdt(mkia_task_t *mtip)
{
	struct idt_gdt_desc_struct cur_desc;
	struct desc_struct *pgdt_table;
	unsigned short tbllimit;
	unsigned long size;

	MKIA_ASSERT(mtip->mti_merge_gdtp == NULL);
	asm volatile("sgdt %0": "=m" (cur_desc));
	tbllimit = cur_desc.limit;
	size = PAGE_ALIGN(tbllimit + 1);

	if (size != PAGE_SIZE) {
		printk("mkia_alloc_gdt: unexpected GDT size %lx\n", size);
		return;
	}
	pgdt_table = (struct desc_struct *) __get_free_page(GFP_KERNEL);
	if (pgdt_table == NULL) {
		printk("mkia_alloc_gdt: allocation failed!\n");
		return;
	}

	/*
	 * Get copy of the linux GDT
	 */
	memcpy(pgdt_table, cur_desc.table, size);
	mtip->mti_merge_gdtp = pgdt_table;
}

/*
 * mkia_internal_set_gdt_entry
 * 
 * This is only called within mkiXX.c to set up the descriptors for the 
 * newly allocated gdt during initialization, and is called from 
 * mkia_set_gdt_entry below
 */
static inline int
mkia_internal_set_gdt_entry(unsigned short sel, unsigned long *new_entry, mkia_task_t *mtip)
{
	struct idt_gdt_desc_struct cur_desc;
	int i;
	unsigned short tbllimit;
	struct desc_struct *pgdt_table;

	asm volatile("sgdt %0": "=m" (cur_desc));
	tbllimit = cur_desc.limit;
	if (sel < 1 || sel > tbllimit) {
		return EINVAL;
	}
	pgdt_table = mtip->mti_merge_gdtp;
	if (sel >= PAGE_SIZE) {
		return EBUSY;
	}
	if (sel >= __KERNEL_CS) {
		return EBUSY;
	}

	/* selector -> index */
	i = sel >> 3;
	pgdt_table[i] = *(struct desc_struct *) new_entry;
	return 0;
}

static inline
void
mkia_setup_ldt_descriptor(mkia_task_t *mtip, void * pldt_table, int limit)
{
	struct desc_struct ldtdesc;
	mtip->mti_current_ldtp = pldt_table;
	mtip->mti_current_ldt_limit = limit;
	ldtdesc.a = (0xffff0000 & ((unsigned long) pldt_table << 16)) | 
	    (0x0000ffff & (limit));
	ldtdesc.b = (0xff000000 & (unsigned long) pldt_table) |
	    (0x000f0000 & limit) | (0x00008200) | 
	    (0x000000ff & ((unsigned long) pldt_table >> 16));
	if (mtip->mti_merge_gdtp == NULL) {
		printk("mki error: gdtp not allocated and "
		    "mkia_setup_ldt_descriptor called\n");
	} else {
		mkia_internal_set_gdt_entry(MKI_LDT_DESC,
			(unsigned long *) &ldtdesc, mtip);
	}
}

static inline 
void
mkia_alloc_ldt(mkia_task_t *mtip)
{
	struct desc_struct *pldt_table;
	char *copydest;
	mm_context_t *contextp;
	void *orig_ldt;
	unsigned short limit;
	int orig_ldt_size;
	int new_ldt_size;

	MKIA_ASSERT(mtip->mti_merge_ldtp == NULL);
	mtip->mti_save_ldtp = NULL;
	mtip->mti_save_ldt_size = 0;

	limit = (LDT_ENTRIES * LDT_ENTRY_SIZE) - 1;
	new_ldt_size = PAGE_ALIGN(limit + 1);
	if (!(pldt_table = vmalloc(new_ldt_size))) {
		printk("mkia_alloc_ldt: vamlloc failed\n");
		return;
	}
	mtip->mti_merge_ldtp = pldt_table;

	contextp = &current->mm->context;
	orig_ldt = contextp->ldt;
	orig_ldt_size = (contextp->size) * LDT_ENTRY_SIZE;

	if (orig_ldt_size) {
		memcpy(pldt_table, orig_ldt, orig_ldt_size);
	}
	if (orig_ldt_size < new_ldt_size) {
		copydest = ((char *) pldt_table) + orig_ldt_size;
		memset(copydest, 0, new_ldt_size - orig_ldt_size);
	}
	mkia_setup_ldt_descriptor(mtip, pldt_table, limit);
}

/*
 * mkia_load_mki_descriptors
 *
 * switch in the task's gdt, idt, ldt, and tr
 */
static inline void
mkia_load_mki_descriptors(mkia_task_t *mtip)
{
	struct idt_gdt_desc_struct cur_desc;
	unsigned int *gdttable;

	MKIA_ASSERT(mtip->mti_flags & MKIF_DESC_ALLOCATED);
	asm volatile("sgdt %0": "=m" (cur_desc));
	cur_desc.table = mtip->mti_merge_gdtp;
	asm volatile("lgdt %0": : "m" (cur_desc));
	asm volatile("lldt %%ax": : "a" (MKI_LDT_DESC));

	/*
	 * Point to second part of TSS descriptor and clear busy
	 * bit prior to ltr
	 */
	gdttable = mtip->mti_merge_gdtp;
	gdttable[(MKI_TSS_ENTRY << 1) + 1] &= (unsigned int) 0xfffffdff;
	asm volatile("ltr %%ax": : "a" (MKI_TSS_DESC));
	if (mkia_idt != NULL) {
		asm volatile("lidt %0": : "m" (mkia_idt_.limit));
	}
}


void
mkia_alloc_descriptors(struct task_struct *curr)
{
	mkia_task_t *mtip=mkia_get_task_info(curr);

	MKIA_ASSERT(!(mtip->mti_flags & MKIF_DESC_ALLOCATED));
	mkia_alloc_gdt(mtip);
	mkia_alloc_ldt(mtip);
	mkia_alloc_tss(curr, MKI_TSS_IOBITMAP_BITS);
	MKIA_ASSERT(mtip->mti_merge_gdtp != NULL);
	MKIA_ASSERT(mtip->mti_merge_ldtp != NULL);
	MKIA_ASSERT(mtip->mti_merge_tssp != NULL);
	mtip->mti_flags |= MKIF_DESC_ALLOCATED;
}

/*
 * mkia_cleanup_descriptors()
 *    needs to be called for each MKI process that exits.
 *
 * This function is always called in task context so no locking is
 * necessary.
 */
void
mkia_cleanup_descriptors(mkia_task_t *mtip)
{
	MKIA_ASSERT(mtip->mti_flags & MKIF_DESC_ALLOCATED);
	mkia_free_ldt(mtip);
	mkia_free_tss(mtip);
	mkia_free_gdt(mtip);
	MKIA_ASSERT(mtip->mti_merge_gdtp == NULL);
	MKIA_ASSERT(mtip->mti_merge_ldtp == NULL);
	MKIA_ASSERT(mtip->mti_merge_tssp == NULL);
	mtip->mti_flags &= ~MKIF_DESC_ALLOCATED;
}

/*
 * mkia_free_task_info()
 */
void mkia_free_task_info(mkia_task_t *mtip)
{
	if(mtip->mti_merge_gdtp != NULL) {
		printk("WARNING! freeing MKI task info struct without "
		    "first freeing gdt. Freeing gdt/ldt/tss\n");
		mkia_load_linux_descriptors();
		mkia_cleanup_descriptors(mtip);
	}
	kfree(mtip);
}

/*
 * mhi_switch_to
 */
void
mhia_switch_to(void *arg1_next, void *arg2_unused)
{
	struct task_struct *next = arg1_next;
	mkia_task_t *mtip;


	/*
	 * If the next task is MARKED, switch in our descriptor tables
	 * and call the SWITCH_TO hook don't use mki_get_task_info,
	 * because it allocates it if NULL.
	 */
	if (unlikely(((mtip = next->mki_task_info) != NULL) &&
	    (mtip->mti_flags & MKIF_MARKED))) {
		/* switch in private gdt, idt, ldt and tr */
		mkia_load_mki_descriptors(mtip);
		mtip->mti_flags |= MKIF_IN_SWITCH;
		(void) (*mhia_table[SWITCH_TO])(mtip->mti_vm86p);
		mtip->mti_flags &= ~MKIF_IN_SWITCH;
		next->thread.fs = mtip->mti_fs;
		next->thread.gs = mtip->mti_gs;
	}
}

/*
 * mhi_switch_away
 */
void
mhia_switch_away(void *arg1_prev, void *arg2_unused)
{
	struct task_struct *prev = arg1_prev;
	mkia_task_t *mtip;


	/*
	 * Don't use mki_get_task_info, because it allocates it if NULL,
	 * and we don't need _every_ task having an mtip!
	 */
	if (unlikely(((mtip = prev->mki_task_info) != NULL) &&
	    (mtip->mti_flags & MKIF_MARKED))) {
		int zero = 0;
		asm volatile("movl %%fs,%0":"=m" (*(int *)&mtip->mti_fs));
		asm volatile("movl %%gs,%0":"=m" (*(int *)&mtip->mti_gs));
		asm volatile("movl %0, %%fs": : "m" (zero));
		asm volatile("movl %0, %%gs": : "m" (zero)); 

		mkia_load_linux_descriptors();
		mtip->mti_flags |= MKIF_IN_SWITCH;
		(void) (*mhia_table[SWITCH_AWAY])(mtip->mti_vm86p);
		mtip->mti_flags &= ~MKIF_IN_SWITCH;
	}
}

void
mhia_exit(void *arg1_unused, void *arg2_unused)
{
	/* Call the EXIT hook for MARKED tasks */
	mkia_task_t *mtip;
	struct task_struct *curr;
	curr = current;

	/*
	 * Don't use mki_get_task_info, because it allocates it if NULL
	 */
	if (unlikely((mtip = curr->mki_task_info) != NULL)) {
		if (likely(mtip->mti_flags & MKIF_MARKED)) {
			(void) (*mhia_table[THREAD_EXIT])(mtip->mti_vm86p);
			/* put everything back */
			mkia_load_linux_descriptors();
			if (mtip->mti_flags & MKIF_DESC_ALLOCATED) 
				mkia_cleanup_descriptors(mtip);
		}
		mkia_free_task_info(mtip);
		curr->mki_task_info = NULL;
	}
}

/*
 * mhia_ret_user
 *
 * This routine gets called just before the kernel is going to make
 * a kernel mode to user mode transition.  On 2.6, this only gets
 * called for "marked" tasks (or about to be marked tasks), because
 * only marked tasks have TIF_MKI_RETUSER set.
 */
void
mhia_ret_user(void *arg1_regs, void *arg2_unused)
{
	unsigned long *r0ptr = arg1_regs;
	struct task_struct *curr = current;
	mkia_task_t *mtip;

	if (unlikely((mtip = curr->mki_task_info) == NULL))
		return;

	if (unlikely(!(mtip->mti_flags & MKIF_MARKED)))
		return;

	while (curr->flags & PF_FREEZE) {
#ifdef PF_IOTHREAD
		/* Call with PF_IOTHREAD to flush signals */
		refrigerator(PF_IOTHREAD);
#else
		/* Call with PF_NOFREEZE to flush signals */
		refrigerator(PF_NOFREEZE);
#endif	/* PF_IOTHREAD */
	}
	if (signal_pending(current)) {
		/*
		 * We catch signals here so that the lower layer does
		 * not try to do the Linux DOSEMU vm86 handling or any
		 * other kind of usermode signal handling.  With merge a
		 * signal in the VM86 process is always a reason to exit.
		 */
		__asm__ __volatile__("sti\n");
		do_exit(1);
		/* NORETURN */
	}
	if (mtip->mti_event_pending) {
		mtip->mti_event_pending = 0;
		(void) (*mhia_table[RET_USER])(r0ptr);
	}
	MKIA_ASSERT(mkia_get_preempt_count() == 0);
}


/*
 * mhia_swap
 */
void
mhia_swap(void *arg1_priority, void *arg2_gfp_mask)
{
	int priority;
	int gfp_mask;
	int hard_flag;

	priority = (int) arg1_priority;
	gfp_mask = (int) arg2_gfp_mask;

	/*
	 * A "HARD" swap means get rid of all mappings rather than
	 * just aging them.
	 */
	hard_flag = (priority < 6);
	(void) (*mhia_table[SWAP_PAGES])((void *) hard_flag);
}

/* mkia_pgvfree assumes PAGE_SIZE <= size <= 4M */
/* mkia_pgvfree assumes size is multiple of PAGE_SIZE */
void
mkia_pgvfree(void * addr, unsigned long size)
{
	pgd_t * dir;
	pmd_t * pmd;
	pte_t * pte;
	unsigned long end;
	unsigned long address = (unsigned long) addr;

	MKIA_ASSERT((size >= PAGE_SIZE) && (size < 0x400000) &&
		((size & 0xfff) == 0));

	dir = pgd_offset_k(address);
	pmd = pmd_offset(dir, address);
	pte = pte_offset_kernel(pmd, address);
	address &= ~PGDIR_MASK;
	end = address + size;
	if (end > PGDIR_SIZE) {
		end = PGDIR_SIZE;
	}

	/* skip first page and just clear pte of others */

	pte++;
	address += PAGE_SIZE;
	size -= PAGE_SIZE;
	while (address < end) {
		pte_clear(pte);
		address += PAGE_SIZE;
		size -= PAGE_SIZE;
		pte++;
	}
	if (size) {
		dir++;
		pmd = pmd_offset(dir, address);
		pte = pte_offset_kernel(pmd, address);
		while (size) {
			pte_clear(pte);
			size -= PAGE_SIZE;
			pte++;
		}
	}
	vfree(addr);
}

/*
 * void
 * mkia_post_event(void *cookie)
 *
 * Set's an event pending.  NOTE this routine may be called OUT of
 * context and, in fact, out of an interrupt.  We can get away without
 * a BUS-LOCK prefix because we do an assigment rather than a Read-
 * Modify-Write.
 */
void 
mkia_post_event(void *cookie)
{
	struct task_struct *t;
	mkia_task_t *mtip;
	unsigned long flags;

	flags = 0;
	t = (struct task_struct *) cookie;

	mtip = (mkia_task_t *)(t->mki_task_info);
	mtip->mti_event_pending = 1;
   
#ifdef CONFIG_SMP
	/*
	 * If the task is running on a different CPU, force a reschedule
	 * on that CPU.  This will force that task into the kernel if it
	 * is not already there, and on the way back to user mode,
	 * mti_event_pending will get checked. Only do this if we are not
	 * being called from a SWITCH_TO or SWITCH_AWAY hook. Otherwise
	 * we can get a deadlock, since interrupts could be disabled on
	 * this cpu and we could block on the run_queue lock or the reschedule.
	 */
	if (!(mtip->mti_flags & MKIF_IN_SWITCH) && (num_online_cpus() > 1)) {
		mki_kick_if_running(t);
	}
#endif /* CONFIG_SMP */

}
EXPORT_SYMBOL(mkia_post_event);

int mkia_set_gdt_entry(unsigned short sel, unsigned long *new_entry)
{
	int ret;
	mkia_task_t *mtip;

	mtip = mkia_get_task_info(current);
	if (!(mtip->mti_flags & MKIF_DESC_ALLOCATED)) {
		printk("mki warning: allocating descriptors in "
			"mkia_set_gdt_entry"); 
		mkia_alloc_descriptors(current);
	}

	/*
	 * If this is the first time we are setting a descriptor, be sure
	 * the old copies of the linux descriptors are cleared out and
	 * that current->mm->context.segments is correct
	 */
	if (!(mtip->mti_flags & MKIF_GDT_SELECTOR_ADDED)) {
		/* selectors 0-12 */
		mtip->mti_flags |= MKIF_GDT_SELECTOR_ADDED;
		memset(mtip->mti_merge_gdtp, 0, MKI_CLEAR_GDT_AMOUNT);
	}
	if (sel >= MKI_TSS_DESC) {
		return EBUSY;
	}
	if (!(ret = mkia_internal_set_gdt_entry(sel, new_entry, mtip))) {
		/* Everything went ok, so switch in our private descriptors */
		mkia_load_mki_descriptors(mtip);
	}
	return ret;
}
EXPORT_SYMBOL(mkia_set_gdt_entry);

int mkia_set_ldt_entry(unsigned short sel, unsigned long *new_entry)
{
	struct desc_struct *pldte;
	mkia_task_t *mtip;
	struct task_struct *curr = current;

	mtip = mkia_get_task_info(curr);
	MKIA_ASSERT(mtip->mti_merge_ldtp);

	pldte = (struct desc_struct *)
		    ((char *) mtip->mti_current_ldtp + (sel & ~(0x7)));

	*pldte = *(struct desc_struct *) new_entry;
	return 0;
}
EXPORT_SYMBOL(mkia_set_ldt_entry);

int mkia_check_vm86(void)
{
	return ((mkia_get_task_info(current))->mti_flags & MKIF_MARKED) 
	    ? 1 : 0;
}
EXPORT_SYMBOL(mkia_check_vm86);

static inline
int
mkia_alloc_idt(void)
{
	/*
	 * No private IDT yet. Make private copy of IDT.  For F00F bug
	 * systems allocate two pages.  Make a ro version, which points
	 * to the rw.
	 */
	if (boot_cpu_data.f00f_bug) {
		pte_t * pte;
		pte_t * pte_rw;
		unsigned long page;

		mkia_idt_rw = vmalloc(2*PAGE_SIZE);
		page = (unsigned long) mkia_idt_rw;
		pte_rw = pte_offset_kernel(pmd_offset(pgd_offset(&init_mm,
							  page), page), page);
		page += PAGE_SIZE;
		mkia_idt = (struct desc_struct *) page;
		pte = pte_offset_kernel(pmd_offset(pgd_offset(&init_mm,
						       page), page), page);
		__free_page(pte_page(*pte));
		*pte = *pte_rw;
		pte_modify(*pte, PAGE_KERNEL_RO);
		flush_tlb_all();
	}
	else {
		mkia_idt = mkia_idt_rw = (struct desc_struct *)
			__get_free_page(GFP_KERNEL);
	}
	if (mkia_idt == NULL) {
		return 0;
	}
	memcpy(mkia_idt_rw, idt_table, PAGE_SIZE);
	return 1;
}

void mkia_set_idt_entry(unsigned short vect_num, unsigned long *new_entry, unsigned long *prev_entry)
{
	if (mkia_idt == NULL) {
		printk("mkia_set_idt_entry: FAILED - no private IDT\n");
		return;
    	}
	lock_kernel();
	*(struct desc_struct *) prev_entry  = mkia_idt[vect_num];
	mkia_idt_rw[vect_num] = *(struct desc_struct *) new_entry;
	unlock_kernel();
}
EXPORT_SYMBOL(mkia_set_idt_entry);

void mkia_set_idt_dpl(void)
{
	int i;
	struct desc_struct *p;

	/*
	 * Go make all IDT descriptors DPL 0.  Note that Merge
	 * has special case code to enable Linux's "int 0x80"
	 * system calls.
	 */
	lock_kernel();
	p = mkia_idt_rw;
	for (i = 0; i < 256; i++, p++) {
		if (p->a | p->b) {
			p->b &= ~(3 << 13);
		}
	}
	unlock_kernel();
}
EXPORT_SYMBOL(mkia_set_idt_dpl);

void *
mkia_getparm(int request, void *parm)
{
	extern int mkia_disable_kmalloc;
	mkia_task_t *mtip;
	switch (request) {
	case PARM_POST_COOKIE:
		*(void **)parm = current;
		break;
	case PARM_FRAME_BASE:
		/*
		 * This really is just the very bottom of the
		 * stack.  To turn this into frame ptr, Merge
		 * has to subtract off its idea of the frame size.
		 */
		*(void **)parm = ((char *)current_thread_info()) +
			THREAD_SIZE;
		break;
	case PARM_TASK_MASK:
		/*
		 * Since 2.6 requires 5.5.x of merge, just return 0
		 * 5.5.x will use PARM_TASK_MASK_V2.
		 */
		*(unsigned long *)parm = 0;
		break;
	case PARM_TSSP:
		/* This should not be called in 2.4, 
		   mkia_adjust_esp0 is used instead */
		*(void **) parm = NULL;
		break;
	case PARM_PRIVATE:
		*(void **) parm = NULL;
		break;
	case PARM_LDTP:
		if ((mtip = current->mki_task_info) == NULL)
			*(void **) parm = NULL;
		else
			*(void **) parm = mtip->mti_current_ldtp;
		break;
	case PARM_RUID:
		*(uid_t *) parm = current->uid;
		break;

	/* ========================================================= */
	/* Begin Version 2 MKI calls */

	case PARM_TASK_MASK_V2:
		*(unsigned long *)parm = ~(THREAD_SIZE - 1);
		break;

	case PARM_MKI_VERSION:
		*(unsigned int *)parm = 6;
		break;

	case PARM_NUM_CPUS:
		*(unsigned int *)parm = num_online_cpus();
		break;

	case PARM_MAX_NUMPROCS:
		*(unsigned int *)parm = PID_MAX_LIMIT;
		break;

	case PARM_PREEMPT_ENABLE:
#ifdef CONFIG_PREEMPT
		*(void **) parm = mkia_preempt_enable;
#else
		*(int *) parm = 0;
#endif
		break;

	case PARM_PREEMPT_DISABLE:
#ifdef CONFIG_PREEMPT
		*(void **) parm = mkia_preempt_disable;
#else
		*(int *) parm = 0;
#endif
		break;

	case PARM_PREEMPT_COUNT:
		mkia_disable_kmalloc = 0;
#ifdef CONFIG_PREEMPT
		*(void **) parm = mkia_get_preempt_count;
#else
		*(int *) parm = 0;
#endif
		break;

	case PARM_HZ:
		*(int *) parm = HZ;
		break;

	default:
		printk("mkia_getparm: no support for this request %d\n",
			request);
		*(int *) parm = 0;		/* for sanity */
		break;
	}
	return 0;
}
EXPORT_SYMBOL(mkia_getparm);

int
mkia_mem_lock_unlock(unsigned long start, unsigned long end, int lock)
{
	struct task_struct *tsk;
	struct mm_struct *mm;
	struct vm_area_struct * vma;
	unsigned long curaddr;
	unsigned long curend;
	unsigned long pages;

	tsk = current;
	mm = tsk->mm;

	/*
	 * If we're in an interrupt or have no user
	 * context, we must not take the fault..
	 */
	if (in_atomic() || !mm)
		return -EBUSY;

	down_write(&mm->mmap_sem);

	/* Do page alignment */
	start &= PAGE_MASK;
	end = (end + ~PAGE_MASK) & PAGE_MASK;

	curaddr = start;
	while (curaddr < end) {
		vma = find_vma(mm, curaddr);
		if (!vma) {
			printk("up_and_fail: no vma found! "
			    "(mm=0x%p curaddr=0x%lx)\n", mm, curaddr);
			goto up_and_fail;
		}
		if (vma->vm_start != curaddr) {
			goto up_and_fail;
		}
		if (vma->vm_end > end) {
			goto up_and_fail;
		}
		curaddr = vma->vm_end;
	}

	/*
	 * OK.  Now that we've done that, we can go ahead and lock (or unlock)
	 * the pages.  If we are locking we just have to set VM_LOCKED and
	 * then call make_pages_present to make sure the ptes get filled in.
	 * VM_LOCKED will prevent swap_out_vma() from stealing the pages.
	 * For unlock, we just have to clear VM_LOCKED.
	 */
	curaddr = start;
	while (curaddr < end) {
		vma = find_vma(mm, curaddr);
		curend = vma->vm_end;
		pages = (curend - curaddr) >> PAGE_SHIFT; 
		if (lock) {
			MKIA_ASSERT((vma->vm_flags & VM_LOCKED) == 0);
			vma->vm_flags |= VM_LOCKED;
			mm->locked_vm += pages;
			make_pages_present(curaddr, curend);
		} else {
			MKIA_ASSERT(vma->vm_flags & VM_LOCKED);
			vma->vm_flags &= ~VM_LOCKED;
			MKIA_ASSERT(mm->locked_vm >= pages);
			mm->locked_vm -= pages;
		}
		curaddr = curend;
	}
	up_write(&mm->mmap_sem);
	return 0;

up_and_fail:
	up_write(&mm->mmap_sem);
	printk("_mki_mem_lock_unlock: bad params: "
	    "start %lx, end %lx, lock %x\n", start, end, lock);
	return -EINVAL;
}

int
mkia_set_private_ldt(void *ldtp, size_t limit)
{
	unsigned long ldtaddr;
	unsigned long endaddr;
	struct task_struct *curr;
	mm_context_t *contextp;
	mkia_task_t *mtip;

	curr=current;
	mtip = mkia_get_task_info(curr);
	MKIA_ASSERT((mtip->mti_flags & MKIF_MARKED) != 0);
	if (curr->mm == NULL)
		return EINVAL;

	contextp = &curr->mm->context;
	if (! (mtip->mti_flags & MKIF_SETLDT_DONE)) {
		mtip->mti_save_ldtp = mtip->mti_current_ldtp;
		mtip->mti_save_ldt_size = mtip->mti_current_ldt_limit + 1;
		mtip->mti_flags |= MKIF_SETLDT_DONE;
		if (ldtp == NULL)
			return 0;
	}
	if (ldtp == NULL) {
		/* NULL means restore the saved original LDT */
		ldtp = mtip->mti_save_ldtp;
		limit = (mtip->mti_save_ldt_size * LDT_ENTRY_SIZE) - 1;
	}

	/* Unlock previous LDT */
	ldtaddr = (unsigned long) mtip->mti_current_ldtp;
	if (ldtaddr < MKI_END_USER_ADDR) {
		endaddr = ldtaddr +
			mtip->mti_current_ldt_limit + 1;
		mkia_mem_lock_unlock(ldtaddr, endaddr, 0 /* Unlock */);
	}

	/* Lock the new LDT */
	ldtaddr = (unsigned long) ldtp;
	if (ldtaddr < MKI_END_USER_ADDR) {
		endaddr = ldtaddr + limit + 1;
		mkia_mem_lock_unlock(ldtaddr, endaddr, 1 /* Lock */);
	}

	mkia_setup_ldt_descriptor(mtip, ldtp, limit);
	asm volatile("lldt %%ax": : "a" (MKI_LDT_DESC));
	return 0;
}
EXPORT_SYMBOL(mkia_set_private_ldt);

void mkia_pgfault_get_state(int *pfault_ok, void *fcstate_arg)
{
	mki_fault_catch_t *fcstate = fcstate_arg;
	unsigned long faulting_eip;

	/*
	 *  Context check.  Make sure that we are not in the middle of
	 *  servicing an interrupt.  If so, then the fault catch information
	 *  is not valid.
	 */
	if (in_interrupt()) {
		*pfault_ok = 0;
		return;
	}

	/*
	 *  Save the old state and clear the current.
	 */
	*pfault_ok = 1;
	faulting_eip = (unsigned long) 
		fcstate->mkifc_os_dependent[FC_FAULT_EIP];

	fcstate->mkifc_catching_user_fault = 
		(search_exception_tables(faulting_eip) ? 1 : 0);

	fcstate->mkifc_os_dependent[FC_SAVE_FS] = (int) (get_fs()).seg;
	set_fs(KERNEL_DS);
}
EXPORT_SYMBOL(mkia_pgfault_get_state);

void mkia_pgfault_restore_state(void *fcstate_arg)
{
	mki_fault_catch_t *fcstate = fcstate_arg;
	set_fs(MAKE_MM_SEG(fcstate->mkifc_os_dependent[FC_SAVE_FS]));
	fcstate->mkifc_os_dependent[FC_SAVE_FS] = 0;
}
EXPORT_SYMBOL(mkia_pgfault_restore_state);

/*
 * void *
 *	mkia_alloc_priv_tss(void)
 */
void *
mkia_alloc_priv_tss(void)
{
	struct tss_struct *t;
	mkia_task_t *mtip;

	mtip = mkia_get_task_info(current);
	if (mtip->mti_vm86p == NULL) {
		printk("mki-adapter: current task has no vm86p on "
		    "entry to mki_alloc_priv_tss\n");
		return NULL;
	}
	
	MKIA_ASSERT(mtip->mti_flags & MKIF_DESC_ALLOCATED);

	/* Just in case, if we have not done so already, alloc the tss 
	 * and make sure the iobitmap is properly initialized.
	 */
	if (!(mtip->mti_flags & MKIF_DESC_ALLOCATED)) {
		printk("mki-adapter: warning! allocating descriptors "
		    "from alloc_priv_tss\n");
		mkia_alloc_descriptors(current);
	}
	t = mtip->mti_merge_tssp;

	return (void *) t;
}
EXPORT_SYMBOL(mkia_alloc_priv_tss);

void mkia_set_vm86p(void *vm86p)
{
	struct task_struct *curr=current;
	(mkia_get_task_info(curr))->mti_vm86p = vm86p;
}
EXPORT_SYMBOL(mkia_set_vm86p);

void *
mkia_get_vm86p(void)
{
	return (mkia_get_task_info(current))->mti_vm86p;
}
EXPORT_SYMBOL(mkia_get_vm86p);

void mkia_mark_vm86(void)
{
	mkia_task_t *mtip;
	mtip = mkia_get_task_info(current);
	mtip->mti_flags |= MKIF_MARKED;
	set_thread_flag(TIF_MKI_RETUSER);

	/* Ok, now switch in the mki descriptors */	
	if (!(mtip->mti_flags & MKIF_DESC_ALLOCATED)) {
		printk("mki-adapter: warning! allocating descriptors "
		    "from mki_mark_vm86\n");
		mkia_alloc_descriptors(current);
	}
	mkia_load_mki_descriptors(mtip);
}
EXPORT_SYMBOL(mkia_mark_vm86);

void mkia_yield(void)
{
	/*
	 * Modules are supposed to just call yield() instead of
	 * sys_sched_yield().
	 */
	yield();
}
EXPORT_SYMBOL(mkia_yield);

void mkia_clear_vm86(void)
{
	mkia_task_t *mtip;

	mtip = mkia_get_task_info(current);

	/*
	 * Make sure we release our descriptors and go back to the 
	 * linux gdt, idt, tss, and ldt
	 */
	mkia_load_linux_descriptors();
	mkia_cleanup_descriptors(mtip);
	
	mtip->mti_flags &= ~MKIF_MARKED;
	clear_thread_flag(TIF_MKI_RETUSER);
	set_thread_flag(TIF_IRET);
}
EXPORT_SYMBOL(mkia_clear_vm86);

int
mkia_ftruncate_k(int fd, off_t length)
{
	return (int) sys_ftruncate(fd, (unsigned long) length);
}
EXPORT_SYMBOL(mkia_ftruncate_k);

/*
 * mkia_alloc_task_info()
 */
mkia_task_t * 
mkia_alloc_task_info(void)
{
	mkia_task_t *mtip;

	mtip = (mkia_task_t *) kmalloc(sizeof(mkia_task_t), GFP_KERNEL);
	if (mtip)
		memset(mtip, 0, sizeof(mkia_task_t));
	return mtip;
}

void
mkia_alloc_tss(void *cookie, int iob_size)
{
	struct desc_struct tssdesc;
	unsigned short tbllimit;
	int size;
	struct tss_struct *ptss_table;
	struct tss_struct *oldtss;
	mkia_task_t *mtip = mkia_get_task_info((struct task_struct *) cookie);

	MKIA_ASSERT(mtip->mti_merge_tssp == NULL);
	if ((iob_size < 0) ||
	    (iob_size > 0x0ffff)) {
		printk("mkia_alloc_tss: iob_size of 0x%x ignored\n", iob_size);
		return;
	}
	oldtss = &per_cpu(init_tss, smp_processor_id());
	tbllimit = MKI_TSS_LIMIT(iob_size);
	size = PAGE_ALIGN(tbllimit + 1);
	ptss_table = kmalloc(size,GFP_KERNEL);
	MKIA_ASSERT(ptss_table != NULL);

	/* Make sure that the iobitmap is properly initialized */
	memcpy(ptss_table, oldtss, size);
	ptss_table->ldt = (unsigned short) MKI_LDT_ENTRY;
	ptss_table->io_bitmap_base = IO_BITMAP_OFFSET;
	memset(ptss_table->io_bitmap, 0xff, (iob_size >> 3) + 1);
	mtip->mti_merge_tssp = ptss_table;
	tssdesc.a = (0xffff0000 & ((unsigned long) ptss_table << 16)) | 
	  (0x0000ffff & (tbllimit));
	tssdesc.b = (0xff000000 & (unsigned long) ptss_table) |
	  (0x000f0000 & tbllimit) | (0x00008900) | 
	  (0x000000ff & ((unsigned long) ptss_table >> 16));

	if (mtip->mti_merge_gdtp == NULL) {
		printk("mki error: gdtp not allocated and "
		    "mkia_alloc_tss called\n");
	} else {
		mkia_internal_set_gdt_entry(MKI_TSS_DESC,
		    (unsigned long *) &tssdesc, mtip);
	}
}

#if defined(CONFIG_KDB)
#include <linux/kdb.h>
int mkia_assert_debugger = 1;
#else
int mkia_assert_debugger = 0;
#endif

/*
 * void
 *	mkia_enter_debugger(int reason, int error, struct pt_regs *regs)
 */
void
mkia_enter_debugger(int reason, int error, void *regs_arg)
{
#ifdef CONFIG_KDB
	struct pt_regs *regs = regs_arg;

	/*
	 * Enter the debugger that is currently in use, if any.
	 */
	switch (reason) {
	case 1:		/* Called with error and regs values */
		kdb(KDB_REASON_DEBUG, error, regs);
		break;
	default:
		/*FALLSTHRU*/
	case 0:		/* Called from an "assert" or some other place that
			 * does not have an error code or regs associated.
			 */
		if (in_interrupt()) {
			kdb(KDB_REASON_CALL, 0, 0);
		} else {
			KDB_ENTER();
		}
		break;
	}
#else
	(void) printk("mkia_enter_debugger: no debugger available\n"); 
#endif
}
EXPORT_SYMBOL(mkia_enter_debugger);

void
mkia_assert_failure(char *exprstr, char *filename, int linenum)
{
	/*
	 * Note that we make this a separate function so that
	 * we can put a breakpoint here to trace the problems.
	 */
	mkia_preempt_disable();
	(void) printk(KERN_ERR
	    "MKI Assertion \"%s\" failed: file \"%s\", line %d\n", 
		    exprstr, filename, linenum);
	if (mkia_assert_debugger) {
		mkia_enter_debugger(0, 0, 0);
	} else {
		panic("MKI Assertion \"%s\" failed: file \"%s\", line %d\n", 
		    exprstr, filename, linenum);
	}
	mkia_preempt_enable_nosched();
}

mkia_64bit_t 
mkia_virt_to_phys(void *kaddr)
{
	unsigned long vaddr;
	pte_t *pte;
	pte_t entry;

	if (kaddr < high_memory)
		return virt_to_phys(kaddr);

	vaddr = (unsigned long) kaddr;
	pte = pte_offset_kernel(pmd_offset(pgd_offset_k(vaddr), vaddr), vaddr);
	if (pte == NULL)
		return (mkia_64bit_t) -1;
		
	entry = *pte;
	if (! pte_present(entry))
		return (mkia_64bit_t) -1;

    	return (mkia_64bit_t) (pte_val(entry) & PAGE_MASK);
}
EXPORT_SYMBOL(mkia_virt_to_phys);

mkia_64bit_t *
mkia_get_pagedir(void)
{
	struct mm_struct *mm;

	mm = current->mm;
	return (mkia_64bit_t *) mm->pgd;
}
EXPORT_SYMBOL(mkia_get_pagedir);

unsigned long
mkia_mmap_k(unsigned long addr, size_t len, int prot, int flags,
   int filedes, off_t offset, int *errnop)
{
	struct file * file = NULL;
	int error = -EFAULT;

	if ((prot & PROT_USER) == 0) {
		printk("mmap_k: Non-USER mapping requested\n");
	}

	if (!(flags & MAP_ANONYMOUS)) {
		error = -EBADF;
		file = fget(filedes);
		if (!file)
			goto out;
	}
	flags &= ~(MAP_EXECUTABLE | MAP_DENYWRITE);

	down_write(&current->mm->mmap_sem);
	error = do_mmap(file, addr, len, prot, flags, offset);
	if (file)
		fput(file);
	up_write(&current->mm->mmap_sem);

out:
	if ((error < 0) && (error > -4096)) {
		*errnop = error;
		error = -1;
	} else {
		*errnop = 0;
	}
	return ((unsigned long) error);
}
EXPORT_SYMBOL(mkia_mmap_k);

int
mkia_munmap_k(unsigned long addr, size_t len)
{
	return sys_munmap(addr, len);
}
EXPORT_SYMBOL(mkia_munmap_k);

unsigned long
mkia_mprotect_k(unsigned long addr, size_t len, int prot)
{
	return sys_mprotect(addr, len, prot);
}
EXPORT_SYMBOL(mkia_mprotect_k);

void * 
mkia_get_file_cookie(int filedes)
{
	struct file *file;
	
	file = fget(filedes);
	return file;
}
EXPORT_SYMBOL(mkia_get_file_cookie);

void mkia_put_file_cookie(void *farg)
{
	struct file *file;
	
	if ((file = (struct file *) farg) != NULL)
		fput(file);
}
EXPORT_SYMBOL(mkia_put_file_cookie);

mkia_64bit_t mkia_add_page_ref(unsigned long vaddr)
{
	struct vm_area_struct *vma;
	mkia_64bit_t retval;
	struct mm_struct *mm;
	int page_pres;
	int tries;
	pgd_t *pgd;
	pmd_t *pmd;
	pte_t *ptep;
	pte_t entry;

	vaddr &= ~(PAGE_SIZE - 1);
	mm = current->mm;
	retval = 0;

	down_write(&mm->mmap_sem);
	vma = find_vma(mm, vaddr);
	if (!vma) {
		printk("mkia_add_page_ref: find_vma failed for %lx\n", vaddr);
	        goto mki_add_page_ref_done;
	}

	entry = __pte(0);
	page_pres = 0;
	ptep = NULL;
	for (tries = 0; tries < 2; tries++) {
		spin_lock(&mm->page_table_lock);
		pgd = pgd_offset(vma->vm_mm, vaddr);
		pmd = pmd_alloc(mm, pgd, vaddr);
		if (pmd && ((ptep = pte_alloc_map(mm, pmd, vaddr)) != 0)) {
			entry = *ptep;
			page_pres = (pte_present(entry) && pte_write(entry));
			if (page_pres) {
				/* break with the page_table_lock held! */
				break;
			}
		}
		spin_unlock(&mm->page_table_lock);
		if (ptep) {
			pte_unmap(ptep);
			ptep = NULL;
		}
		make_pages_present(vaddr, vaddr + PAGE_SIZE);
	}
	if (! page_pres) {
		printk("mkia_add_page_ref: couldn't make %lx present\n", vaddr);
	        goto mki_add_page_ref_done;
	}
	MKIA_ASSERT(pfn_valid(pte_pfn(entry)));
	get_page(pte_page(entry));
	spin_unlock(&mm->page_table_lock);
	if (ptep) {
		pte_unmap(ptep);
		ptep = NULL;
	}
	atomic_inc(&mkia_context_rss);
	retval = pte_val(entry);

mki_add_page_ref_done:
	up_write(&mm->mmap_sem);
	return retval;
}
EXPORT_SYMBOL(mkia_add_page_ref);

int mkia_remove_page_ref(unsigned long vaddr, mkia_64bit_t physpte,
	void *file_cookie, off_t offset)
{
	struct page *pageinfop;
	unsigned long physaddr;
	struct file *file;
	int retcode;
	pte_t pte;

	retcode = 0;
	physaddr = physpte & PAGE_MASK;
	offset &= PAGE_MASK;
	file = (struct file *) file_cookie;
	MKIA_ASSERT(file != NULL);
	if (file == NULL) {
		/* Just in case: handle it gracefully */
		goto rem_page_ref_done;
	}

	pte = *((pte_t *)&physpte);
	if (!pfn_valid(pte_pfn(pte)))
	        goto rem_page_ref_done;

	pageinfop = pte_page(pte);
	if (PageReserved(pageinfop)) {
		mkia_cnt_rpr_pagereserved++;
		goto rem_page_ref_free;
	}

	MKIA_ASSERT(pageinfop->mapping->host != NULL);
	MKIA_ASSERT(file->f_dentry->d_inode == pageinfop->mapping->host);
	MKIA_ASSERT(pageinfop->index == (unsigned long)(offset >> PAGE_SHIFT));

	if ((physpte & (_PAGE_DIRTY | _PAGE_ACCESSED)) == 0) {
		/*
		 * If this is a clean mapping, i.e. no mod or acc bits,
		 * then we can just decrement the page use counts and
		 * be done with it!
		 */
		mkia_cnt_rpr_not_dirty_acc++;
	}
	if (physpte & _PAGE_ACCESSED) {
		mkia_cnt_rpr_accessed++;
		flush_tlb_page(find_vma(current->mm,vaddr), vaddr);
	}
	if (physpte & _PAGE_DIRTY) {
		mkia_cnt_rpr_dirty++;
		set_page_dirty(pageinfop);
	}

rem_page_ref_free:
	MKIA_ASSERT(page_count(pageinfop) >= 2);
	put_page(pageinfop);
	atomic_dec(&mkia_context_rss);
	MKIA_ASSERT(atomic_read(&mkia_context_rss) >= 0);

rem_page_ref_done:
	return retcode;
}
EXPORT_SYMBOL(mkia_remove_page_ref);

/*
 * int
 * mki_uaddr_mapped(unsigned long address)
 *
 * returns:
 *	0 if the address is not mapped
 *	1 if the address is mapped
 */
int
mkia_uaddr_mapped(unsigned long address)
{
	struct mm_struct *mm;
	struct vm_area_struct * vma;

	mm = current->mm;

	down_read(&mm->mmap_sem);
	vma = find_vma(mm, address);
	up_read(&mm->mmap_sem);

	return (vma != NULL);
}

/*
 * int
 * mki_upageflt(unsigned long address, int error_code)
 *
 * returns:
 *	0 if the pagefault could not be resolved
 *	1 if the pagefault could be resolved
 */
int
mkia_upageflt(unsigned long address, int error_code)
{
	struct task_struct *tsk;
	struct mm_struct *mm;
	struct vm_area_struct * vma;
	int write;

	tsk = current;
	mm = tsk->mm;

	/*
	 * If we're in an interrupt or have no user
	 * context, we must not take the fault..
	 */
	if (in_atomic() || !mm)
		goto return_fail;

	down_read(&mm->mmap_sem);

	vma = find_vma(mm, address);
	if (!vma)
		goto up_and_fail;
	if (vma->vm_start <= address)
		goto good_area;
	if (!(vma->vm_flags & VM_GROWSDOWN))
		goto up_and_fail;

	/*
	 * The do_page_fault() code handles stack growth here but
	 * we don't do this because we don't have a standard process
	 * whose stack we can grow.  The emulator has a fixed sized
	 * stack and it's not our job to grow the windows stacks.
	 */

good_area:
	write = 0;
	switch (error_code & 3) {
		default:	/* 3: write, present */
			/* fall through */
		case 2:		/* write, not present */
			if (!(vma->vm_flags & VM_WRITE))
				goto up_and_fail;
			write++;
			break;
		case 1:		/* read, present */
			goto up_and_fail;
		case 0:		/* read, not present */
			if (!(vma->vm_flags & (VM_READ | VM_EXEC)))
				goto up_and_fail;
	}


	switch (handle_mm_fault(mm, vma, address, write)) {
	case VM_FAULT_MINOR:
		tsk->min_flt++;
		break;
	case VM_FAULT_MAJOR:
		tsk->maj_flt++;
		break;
	default:
		goto up_and_fail;
	}

	up_read(&mm->mmap_sem);
	return 1;

up_and_fail:
	up_read(&mm->mmap_sem);
	return 0;

return_fail:
	return 0;
}
EXPORT_SYMBOL(mkia_upageflt);

void
mkia_adjust_esp0(int numlongs)
{
	struct tss_struct *tss;
	struct task_struct *curr=current;
	mkia_task_t *mtip;

	tss = &per_cpu(init_tss, get_cpu());
	tss->esp0 -= (numlongs * sizeof(unsigned long));
	curr->thread.esp0 = tss->esp0;
	put_cpu();

	set_thread_flag(TIF_MKI_RETUSER);

	mtip = mkia_get_task_info(curr);

	/* This should always be path where we do the
	 * mkia_alloc_descriptors()
	 */
	if (likely(!(mtip->mti_flags & MKIF_DESC_ALLOCATED))) {
		mkia_alloc_descriptors(curr);
	}

}
EXPORT_SYMBOL(mkia_adjust_esp0);

int
mkia_get_current_task_index(void)
{
	return current->pid;
}


/*
 * void mkia_wakeup()
 *
 * Synchronization variable wakeup.
 */
void
mkia_wake_up(void *wqp)
{
	wake_up((wait_queue_head_t *) wqp);
}
EXPORT_SYMBOL(mkia_wake_up);


/*
 * int mkia_call_svwait()
 *
 * Synchronization variable wait.
 *
 * This must be implemented on the mki side of the world because
 * write_lock_irqsave() compiles differently if __SMP__ is defined.
 * The trick here (and why we can't just use sleep_on() or
 * interruptible_sleep_on()) is that we need to give up the lock
 * AFTER we have set the task state and added ourselves to the wait
 * queue.
 *
 * This function is called indirectly (its address is obtained via
 * mkia_getparm().  This is so that a version 2 win4lin can still run
 * with a version 1 mki, even though version 1 mki does not support
 * a mkia_sv_wait() function.
 */
int
mkia_call_svwait(void *wq, volatile int *lockp, int interruptible)
{
	struct task_struct *taskp;
	wait_queue_t wait;
	unsigned long wqflags;
	int retval;

#ifdef CONFIG_PREEMPT
	MKIA_ASSERT(mkia_get_preempt_count() != 0) ;
#endif
	taskp = current;
	init_waitqueue_entry(&wait, taskp);
	if (interruptible) {
		taskp->state = TASK_INTERRUPTIBLE;
	} else {
		taskp->state = TASK_UNINTERRUPTIBLE;
	}

	/* SLEEP_ON_HEAD start */
	spin_lock_irqsave(&((wait_queue_head_t *)wq)->lock, wqflags);
	__add_wait_queue((wait_queue_head_t *)wq, &wait);
	spin_unlock(&((wait_queue_head_t *) wq)->lock);
	/* SLEEP_ON_HEAD end */

	*lockp = 0;     /* Unlock */
	mkia_preempt_enable_nosched();
	schedule();

	/* SLEEP_ON_TAIL start */
	spin_lock_irq(&((wait_queue_head_t *) wq)->lock);
	__remove_wait_queue((wait_queue_head_t *) wq, &wait);
	spin_unlock_irqrestore(&((wait_queue_head_t *) wq)->lock, wqflags);
	/* SLEEP_ON_TAIL end */

	taskp->state = TASK_RUNNING;

	while (taskp->flags & PF_FREEZE) {
#ifdef PF_IOTHREAD
		/* Call with PF_IOTHREAD to flush signals */
		refrigerator(PF_IOTHREAD);
#else
		/* Call with PF_NOFREEZE to flush signals */
		refrigerator(PF_NOFREEZE);
#endif	/* PF_IOTHREAD */
	}

	if (interruptible) {
		retval = (signal_pending(taskp)) ? 1 : 0;
	} else {
		retval = 0;
	}
	return retval;
}
EXPORT_SYMBOL(mkia_call_svwait);

int
mkia_install_hook(int id, int (*hook_fn)(void *))
{
	if ((id >= 0) && (id < NUM_HOOKS)) {
		mhia_table[id] = hook_fn;
		if (id == SWITCH_TO) {
			/*
			 * If we are removing the SWITCH_TO hook, then
			 * merge is loading so make sure we have a
			 * private IDT
			 */
			if (! mkia_alloc_idt())
				return -1;
		}
		return 0;
	}
	return -1;
}
EXPORT_SYMBOL(mkia_install_hook);

void
mkia_remove_hook(int id)
{
	/* 
	 * For now all the dummy hooks return the same value.
	 * If we ever add hooks where mhia_void() is not appropriate
	 * we need to change the code below to a switch() {} statement
	 */
	MKIA_ASSERT(mhia_table != NULL);
	if ((id >= 0) && (id < NUM_HOOKS)) {
		mhia_table[id] = mhia_void;
		if (id == SWITCH_TO) {
			/*
			 * If we are removing the SWITCH_TO hook, then
			 * merge is unloading so clean up the IDT as well.
			 */
			mkia_cleanup_idt();
		}
	}
}
EXPORT_SYMBOL(mkia_remove_hook);
