/*
 ****************************************************************************
 * Copyright 2001-2003 by NeTraverse, Inc.
 * This software is distributed under the terms of the GPL
 * which is supplied in the LICENSE file with this distribution
 ***************************************************************************
 * $Id: mkifunc.h,v 1.2 2003/11/01 00:40:45 rwb Exp $
 ***************************************************************************
 * Function declarations and constants of the Linux specific Merge/Kernel 
 * Interface (mki).
 */

#ifndef MKIFUNC_H
#define MKIFUNC_H

/* take care of some types if not defined. */
#ifndef _LINUX_TYPES_H
#include <sys/types.h>
#endif

typedef unsigned long long mkia_64bit_t;

#define MKI_END_USER_ADDR 0xC0000000

/* mki_getparm() parameter identifiers  */
#define PARM_POST_COOKIE     0  /* pointer to current LWP (lwp_t) */
#define PARM_FRAME_BASE      1  /* stack frame base */
#define PARM_CPU_TYPE        2  /* pointer to CPU type */
#define PARM_PRIVATE         3  /* pointer to the location of a scratch
		                                      memory pointer */
#define PARM_GDTP            4  /* pointer to current GDT for this LWP */
#define PARM_LDTP            5  /* pointer to current LDT for this LWP */
#define PARM_IDTP            6  /* pointer to current IDT for this LWP */
#define PARM_TSSP            7  /* pointer to current TSS for this LWP */
#define PARM_RUID            8  /* real UID for this process */
#define PARM_TASK_MASK       9  /* Old V1 request to get Task Mask */
/* ===== End of MKI Version 1 supported calls ====== */

#define PARM_TASK_MASK_V2   10  /* New V2 request to get the Task Mask, i.e. */
                                /*   the value for esp to get to task struct */
#define PARM_MKI_VERSION    11  /* MKI version number */
#define PARM_NUM_CPUS       12  /* Number of CPUs present */
#define PARM_MAX_NUMPROCS   13  /* Maximum number of tasks */
#define PARM_CURPROC_INDEX  14  /* Index of current task */
#define PARM_SVWAIT_FUNC    15  /* Address of _mki_sv_wait() routine */
#define PARM_SVWAKEUP_FUNC  16  /* Address of _mki_sv_wakeup() routine */
#define PARM_POLLWAKE_FUNC  17  /* Address of _mki_poll_wake() routine */
#define PARM_PREEMPT_ENABLE 18  /* preemtion enable function */
#define PARM_PREEMPT_DISABLE 19 /* preemtion disable function */
#define PARM_PREEMPT_COUNT  20  /* preemtion count function */
#define PARM_HZ             21  /* get HZ value */

/*
 * merge function offset into the hook functions table.
 */
#define SWITCH_AWAY   0
#define SWITCH_TO     1
#define THREAD_EXIT   2
#define RET_USER      3
#define SIGNAL        4
#define QUERY         5
#define SWAP_PAGES    6
#define NUM_HOOKS     7

/*
 *  Index values for the os dependent portion of mki_fault_catch_t
 */
#define MKI_FC_SIZE	6

#define FC_FAULT_EIP	0
#define FC_SAVE_FS	1

typedef struct {
	int mkifc_catching_user_fault; /* Boolean */
	int mkifc_os_dependent[MKI_FC_SIZE]; /* OS dependent state */
} mki_fault_catch_t;

/* MKI version of vaddr_t */
typedef unsigned long mkiul_t;

/* MCV: MKI Caller Version */
#define MCV_MAGIC		(0xfabc0000)
#define MCV_MAGIC_MASK		(0xffff0000)
#define MCV_VERSION_MASK	(0x0000ffff)
#define MCV_MAKE_VER(ver)	(((ver) & MCV_VERSION_MASK) | MCV_MAGIC)

/*
 * mki_enable_context_cookies data type
 * supports fast merge context switch
 * see platinum MKI interface doc for details.
 */
typedef struct mki_ecc {
	mkiul_t 	mkiecc_addr;  /* context base addr */
	size_t		mkiecc_len;   /* context size */
	void *		mkiecc_cookie;/* outarg */
} mki_ecc_t;

#define MKI_TSS_ENTRY (GDT_ENTRY_KERNEL_BASE - 2)
#define MKI_TSS_DESC  (MKI_TSS_ENTRY << 3)
#define MKI_LDT_ENTRY (GDT_ENTRY_KERNEL_BASE - 1)
#define MKI_LDT_DESC  (MKI_LDT_ENTRY << 3)
#define MKI_CLEAR_GDT_AMOUNT MKI_TSS_DESC

/* MKI info structure */
struct mki_task_info {
        void *mti_vm86p;
        void *mti_merge_gdtp;
        void *mti_save_ldtp;
	int mti_save_ldt_size;
        void *mti_merge_ldtp;
        void *mti_merge_tssp;
        void *mti_current_ldtp;
        size_t mti_current_ldt_limit;
        size_t mti_current_tss_limit;
        unsigned char mti_flags;
        unsigned char mti_event_pending;
        int mti_fs;
        int mti_gs;
};

typedef struct mki_task_info mkia_task_t;

/* defines for the mki_task_flags field */
#define MKIF_MARKED             0x01
#define MKIF_SETLDT_DONE        0x02
#define MKIF_DESC_ALLOCATED     0x04
#define MKIF_GDT_SELECTOR_ADDED 0x08
#define MKIF_TASK_CLEARED       0x10
#define MKIF_IN_SWITCH          0x20

#endif /* MKIFUNC_H */
