/*
 ****************************************************************************
 * Copyright 2001-2003 by NeTraverse, Inc.
 * This software is distributed under the terms of the GPL
 * which is supplied in the LICENSE file with this distribution
 ***************************************************************************
 * $Id: timer.c,v 1.2 2003/11/01 00:40:45 rwb Exp $
 ***************************************************************************
 * 
 * Provide unix-like timeout() and untimeout() interfaces.
 */

#include <linux/config.h>
#include <linux/types.h>
#include <linux/ptrace.h>
#include <linux/version.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/module.h>
#include "mkifunc.h"

struct linux_unix_timer {
	struct linux_unix_timer *lut_next;
	struct linux_unix_timer *lut_prev;
        void (*lut_unix_timeout_func)(void *);
	int lut_timeoutid;
        void *lut_unix_timeout_arg;
	struct timer_list lut_linux_timer;
};

struct linux_unix_timer lut_list_head = { &lut_list_head, &lut_list_head } ;
int lut_unique_id = 1;

spinlock_t mkia_lut_lock = SPIN_LOCK_UNLOCKED;
#define DECLARELOCKSAVE(flags) unsigned long flags=0
#define LOCKTIMER(flags) spin_lock_irqsave(&mkia_lut_lock, flags)
#define UNLOCKTIMER(flags) spin_unlock_irqrestore(&mkia_lut_lock, flags)

void
do_mki_timeout(unsigned long timer_arg)
{
	struct linux_unix_timer *target_lutp;
	struct linux_unix_timer *lutp;
	void (*timeout_func)(void *);
	void *timeout_arg;
	DECLARELOCKSAVE(flags);

	target_lutp = (struct linux_unix_timer *) timer_arg;
	timeout_func = target_lutp->lut_unix_timeout_func;
	timeout_arg = target_lutp->lut_unix_timeout_arg;

	/* See if the element is still on the active list */
	LOCKTIMER(flags);
	for (lutp = lut_list_head.lut_next; lutp != &lut_list_head;
	     lutp = lutp->lut_next) {
	     	if (lutp == target_lutp)
			break;
	}
	if (lutp != &lut_list_head) {
		lutp->lut_next->lut_prev = lutp->lut_prev;
		lutp->lut_prev->lut_next = lutp->lut_next;
		UNLOCKTIMER(flags);

		del_timer(&lutp->lut_linux_timer);
		kfree(lutp);
		(*timeout_func)(timeout_arg);
	} else {
		UNLOCKTIMER(flags);
		printk("Merge: mki-adapter: not doing timeout func\n");
		printk("Merge: mki-adapter: func %p, arg %p\n",
			timeout_func, timeout_arg);
		(*timeout_func)(timeout_arg);
	}
}

int
timeout(void (*timeout_func)(void *), void *timeout_arg, long ticks)
{
	struct linux_unix_timer *lutp;
	DECLARELOCKSAVE(flags);

	lutp = kmalloc(sizeof(*lutp), GFP_ATOMIC);
	if (lutp == NULL) {
		printk("Merge: timeout: kmalloc failed\n");
		return 0;
	}
	memset(lutp, 0, sizeof(*lutp));
	init_timer(&lutp->lut_linux_timer);
	lutp->lut_linux_timer.expires = jiffies + ticks;
	lutp->lut_linux_timer.data = (unsigned long) lutp;
	lutp->lut_linux_timer.function =  do_mki_timeout;
	lutp->lut_unix_timeout_func = timeout_func;
	lutp->lut_unix_timeout_arg = timeout_arg;

	LOCKTIMER(flags);
	lutp->lut_timeoutid = lut_unique_id++;
	lutp->lut_next = &lut_list_head;
	lutp->lut_prev = lut_list_head.lut_prev;
	lut_list_head.lut_prev->lut_next = lutp;
	lut_list_head.lut_prev = lutp;
	UNLOCKTIMER(flags);

	add_timer(&lutp->lut_linux_timer);

	return (lutp->lut_timeoutid);
}
EXPORT_SYMBOL(timeout);

void
untimeout(int id)
{
	struct linux_unix_timer *lutp;
	DECLARELOCKSAVE(flags);

	if (id == 0) 
		return;

	LOCKTIMER(flags);
	for (lutp = lut_list_head.lut_next; lutp != &lut_list_head;
	     lutp = lutp->lut_next) {
	     	if (lutp->lut_timeoutid == id) {
			lutp->lut_next->lut_prev = lutp->lut_prev;
			lutp->lut_prev->lut_next = lutp->lut_next;
			break;
		}
	}
	UNLOCKTIMER(flags);

	if (lutp != &lut_list_head) {
		del_timer(&lutp->lut_linux_timer);
		memset(lutp, 0, sizeof(*lutp));
		kfree(lutp);
	}
}
EXPORT_SYMBOL(untimeout);
