/*
 ****************************************************************************
 * Copyright 2001-2003 by NeTraverse, Inc.
 * This software is distributed under the terms of the GPL
 * which is supplied in the LICENSE file with this distribution
 ***************************************************************************
 * $Id: vnetint-pub.h,v 1.2 2003/11/01 00:40:45 rwb Exp $
 ***************************************************************************
 */

/* Interceptor status */
struct vnetint_status {
        int             flags;
        unsigned        NumPktsConsumed;
        unsigned        NumPktsFromVnet;
        unsigned        NumPktsUp;
        unsigned        NumPktsDown;
        ETH_ADDR_T      PhysMacAddr;
};

#define MAX_INTERCEPTORS 16
#define MAXIFNAMELEN 8
/* Interceptor state flags */
#define VNETINTF_ACTIVE       0x2                /* interceptor is active */
#define VNETINTF_MACSET       0x10               /* we got a mac address */
#define VNETINT_ACT(s) \
        (((s)->status.flags & VNETINTF_ACTIVE))

typedef int (*xmit_func_t)(struct sk_buff *skb, struct net_device *dev);
typedef int (*recv_func_t)(struct sk_buff *skb, struct net_device *dev,
		struct packet_type *pt);
typedef int (*filter_func_t)(unsigned char *ucp, unsigned int len,
		struct sk_buff *skb);

/* Interceptor device */
struct vnetint_ifdev {
        char ifname[MAXIFNAMELEN];
        int count;
        void *dev;
        xmit_func_t real_xmit;
        struct vnetint_ifdev *next;
};

/* Interceptor protocol */
struct vnetint_proto {
        int protocol;
        int count;
        struct packet_type *ptype;
	recv_func_t real_recv;
        struct vnetint_proto *next;
};

/* Interceptor device/protocol pairs */
struct vnetint_pcb {
	int count;
        struct vnetint_status status;
        struct vnetint_ifdev *ifdev;
        struct vnetint_proto *proto;
        filter_func_t filter_func;
        struct vnetint_pcb *next;
};

/* Interceptor filter */
struct vnetint_filter {
        struct vnetint_pcb *handle;
        struct vnetint_status status;
        char ifname[MAXIFNAMELEN];
        int protocol;
        filter_func_t intercept_func;
        void *criteria;
};

/* vnetint_CtrlStatus function codes */
#define INTERCEPTOR_GOACTIVE  1
#define INTERCEPTOR_GOPASSIVE 2
#define INTERCEPTOR_GETSTATS  3
#define INTERCEPTOR_NUMBER    4
#define INTERCEPTOR_SETFILTER 5
/* vnetint_ProcessPacket defines */
#define UPSTREAM   1
#define DOWNSTREAM 0

